#!/usr/bin/env python3
"""Manages project database.

Usage:
    pr-db [options] cluster (create | start | stop | status | drop)
    pr-db [options] user create <username>
    pr-db [options] user list
    pr-db [options] user drop <username>
    pr-db [options] db create <dbname> <username>
    pr-db [options] db shell <dbname>
    pr-db [options] db list
    pr-db [options] db drop <dbname>
    pr-db -h | --help

Options:
    -h --help           Print help
    --version           Print prdb version
    --config=<file>     Custom configuration file
    --debug             Print debug info
"""
import glob
import logging
import os
import shutil
import subprocess
import sys
from configparser import ConfigParser

import psycopg2
from docopt import docopt

__version__ = '0.2'

_LOGGER = logging.getLogger(__name__)


class PrdbError(Exception):
    """An error occured in this script."""


class SystemClusterError(PrdbError):
    """Raised for actions which can't be completed, because system cluster is used."""


AUTO_CONF_APPENDIX = "port = {port}\n" \
                     "unix_socket_directories = '{unix_socket_directories}'\n" \
                     "max_prepared_transactions = {max_prepared_transactions}\n" \
                     "log_min_duration_statement = {log_min_duration_statement}\n" \
                     "log_line_prefix = '{log_line_prefix}'\n" \
                     "fsync = {fsync}\n" \
                     "timezone = {timezone}\n"


def _run_command(cmd, *args, **kwargs):
    """Run `subprocess.run` with logging."""
    _LOGGER.debug("Running command: %s", ' '.join(cmd))
    return subprocess.run(cmd, *args, **kwargs)


def _is_cluster_running(config):
    """Return whether cluster is running."""
    pg_ctl = os.path.join(config.get('cluster', 'pg_bin'), 'pg_ctl')
    cmd = [pg_ctl, 'status', '-D', config.get('cluster', 'root_dir')]
    result = _run_command(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    return not result.returncode


def _get_cursor(dbname, config):
    """Create and return cursor to the database.

    The cursor will be set in autocommit transaction.
    """
    root_dir = config.get('cluster', 'root_dir')
    if not root_dir:
        host = None
        port = None
    else:
        host = os.path.join(root_dir, 'sockets')
        port = config.get('cluster', 'port')
    connection = psycopg2.connect(host=host, port=port, dbname=dbname)
    connection.autocommit = True
    return connection.cursor()


def _execute_sql(dbname, config, sql, args=None):
    """Execute SQL in the target database.

    @return: Cursor with the SQL executed
    @raise PrdbError: If database returned error.
    """
    cursor = _get_cursor(dbname, config)
    try:
        cursor.execute(sql, args)
    except psycopg2.ProgrammingError as error:
        raise PrdbError("Database returned error: {error}".format(error=error))
    return cursor


########################################################################################################################
# Cluster commands


def create_cluster(options, config):
    """Create database cluster."""
    sys.stdout.write("Creating cluster...\n")
    root_dir = config.get('cluster', 'root_dir')

    if os.path.exists(root_dir):
        raise PrdbError("Cluster directory '{}' already exists.".format(root_dir))

    # Create cluster
    initdb = os.path.join(config.get('cluster', 'pg_bin'), 'initdb')
    cmd = [initdb, '--pgdata', root_dir, '--locale', config.get('cluster', 'locale')]
    result = _run_command(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode:
        raise PrdbError("Cluster can't be created.")

    unix_socket_directories = os.path.join(root_dir, 'sockets')
    os.mkdir(unix_socket_directories)

    # Configure cluster
    config_data = {
        'port': config.getint('cluster', 'port'),
        'unix_socket_directories': unix_socket_directories,
        'max_prepared_transactions': config.getint('cluster', 'max_prepared_transactions'),
        'log_min_duration_statement': config.getint('cluster', 'log_min_duration_statement'),
        'log_line_prefix': config.get('cluster', 'log_line_prefix'),
        'fsync': config.get('cluster', 'fsync'),
        'timezone': config.get('cluster', 'timezone'),
    }
    with open(os.path.join(root_dir, 'postgresql.auto.conf'), 'a') as auto_conf:
        auto_conf.write(AUTO_CONF_APPENDIX.format(**config_data))


def start_cluster(options, config):
    """Start database cluster."""
    sys.stdout.write("Starting cluster...\n")
    pg_ctl = os.path.join(config.get('cluster', 'pg_bin'), 'pg_ctl')
    cmd = [pg_ctl, 'start', '-w', '-D', config.get('cluster', 'root_dir')]
    result = _run_command(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode:
        raise PrdbError("Unable to start cluster.")


def stop_cluster(options, config):
    """Stop database cluster."""
    sys.stdout.write("Stopping cluster...\n")
    pg_ctl = os.path.join(config.get('cluster', 'pg_bin'), 'pg_ctl')
    cmd = [pg_ctl, 'stop', '-w', '-D', config.get('cluster', 'root_dir')]
    result = _run_command(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode:
        raise PrdbError("Unable to stop cluster.")


def status_cluster(options, config):
    """Get cluster status."""
    if _is_cluster_running(config):
        sys.stdout.write("Cluster is running\n")
    else:
        raise PrdbError("Cluster is not running.")


def drop_cluster(options, config):
    """Drop database cluster."""
    sys.stdout.write("Deleting cluster...\n")
    root_dir = config.get('cluster', 'root_dir')

    if not os.path.exists(root_dir):
        raise PrdbError("Cluster directory '{}' doesn't exist.".format(root_dir))

    shutil.rmtree(root_dir)


def cluster_dispatch(options, config):
    """Run requested cluster operation."""
    root_dir = config.get('cluster', 'root_dir')
    if not root_dir:
        raise SystemClusterError('Cluster actions are disabled on system cluster.')

    if options['create']:
        create_cluster(options, config)
    elif options['start']:
        start_cluster(options, config)
    elif options['stop']:
        stop_cluster(options, config)
    elif options['drop']:
        drop_cluster(options, config)
    else:
        # Use safe option 'status' as fallback
        assert options['status']
        status_cluster(options, config)


########################################################################################################################
# User commands


def create_user(options, config):
    """Create user."""
    # Username in CREATE USER can't be escaped.
    sql = "CREATE USER {username}".format(username=options['<username>'])
    if config.getboolean('user', 'createdb'):
        sql += ' WITH CREATEDB'
    _execute_sql('template1', config, sql)


def list_users(options, config):
    """Print list of users."""
    cursor = _execute_sql('template1', config, "SELECT usename FROM pg_user ORDER BY usename")
    for record in cursor:
        sys.stdout.write('{record[0]}\n'.format(record=record))


def drop_user(options, config):
    """Drop users."""
    # Username in DROP USER can't be escaped.
    _execute_sql('template1', config,
                 "DROP USER {username}".format(username=options['<username>']))


def user_dispatch(options, config):
    """Run requested user operation."""
    if options['create']:
        create_user(options, config)
    elif options['drop']:
        drop_user(options, config)
    else:
        # Use safe option 'lists' as fallback
        assert options['list']
        list_users(options, config)


########################################################################################################################
# Database commands


def create_database(options, config):
    """Create database."""
    # Database in CREATE DATABASE can't be escaped.
    _execute_sql('template1', config,
                 "CREATE DATABASE {dbname} WITH OWNER %s".format(dbname=options['<dbname>']),
                 (options['<username>'],))


def list_databases(options, config):
    """Print list of databases."""
    cursor = _execute_sql('template1', config, "SELECT datname FROM pg_database ORDER BY datname")
    for record in cursor:
        sys.stdout.write('{record[0]}\n'.format(record=record))


def drop_database(options, config):
    """Drop database."""
    # Username in DROP DATABASE can't be escaped.
    _execute_sql('template1', config,
                 "DROP DATABASE {dbname}".format(dbname=options['<dbname>']))


def run_database_shell(options, config):
    """Run database shell."""
    root_dir = config.get('cluster', 'root_dir')
    if not root_dir:
        psql_options = []
    else:
        host = os.path.join(root_dir, 'sockets')
        psql_options = ['--host', host, '--port', config.get('cluster', 'port')]
    # The 'psql' is actually only a wrapper which sets up some environment for the 'psql' command itself.
    cmd = ['psql'] + psql_options + [options['<dbname>']]
    _run_command(cmd)


def database_dispatch(options, config):
    """Run requested database operation."""
    if options['create']:
        create_database(options, config)
    elif options['drop']:
        drop_database(options, config)
    elif options['shell']:
        run_database_shell(options, config)
    else:
        # Use safe option 'lists' as fallback
        assert options['list']
        list_databases(options, config)


########################################################################################################################
# Program core


def get_config_file(config_option):
    """Return configuration file location.

    @param config_option: '--config' value passed in command line
    @return: File location or None
    """
    if config_option is not None:
        return config_option
    if os.path.exists('.pr-db.conf'):
        return '.pr-db.conf'
    home_config = os.path.expanduser('~/.pr-db.conf')
    if os.path.exists(home_config):
        return home_config
    return None


def get_config(config_option):
    """Return `ConfigParser` instance.

    @param config_option: '--config' value passed in command line
    """
    config_file = get_config_file(config_option)

    # Find postgres binary
    for pg_bin in sorted(glob.glob('/usr/lib/postgresql/*/bin'), reverse=True):
        if os.path.exists(os.path.join(pg_bin, 'initdb')):
            break
    else:
        pg_bin = ''

    if config_file is None:
        root_dir = ''
    else:
        root_dir = os.path.join(os.path.dirname(config_file), '.pr-db')

    # Create defaults
    defaults = {'cluster': {'pg_bin': pg_bin, 'root_dir': root_dir, 'locale': 'cs_CZ.UTF-8',
                            'max_prepared_transactions': 5, 'log_line_prefix': '%%t [%%p]: [%%l-1] user=%%u,db=%%d ',
                            'log_min_duration_statement': 0, 'fsync': 'off', 'timezone': 'UTC', 'port': '10000'},
                'user': {'createdb': True},
                'pr-db': {'debug': False}}

    # Create config
    config = ConfigParser()
    config.read_dict(defaults)
    if config_file:
        config.read(config_file)
    return config


def main():
    """Run the command."""
    options = docopt(__doc__, version=__version__)
    config = get_config(options['--config'])

    if options['--debug'] or config.getboolean('pr-db', 'debug'):
        logging.basicConfig(level=logging.DEBUG)

    if options['cluster']:
        dispatcher = cluster_dispatch
    elif options['user']:
        dispatcher = user_dispatch
    else:
        assert options['db']
        dispatcher = database_dispatch
    try:
        dispatcher(options, config)
    except PrdbError as error:
        sys.stderr.write('{}\n'.format(error))
        quit(1)


if __name__ == '__main__':
    main()
